from nautobot.apps.ui import TemplateExtension

from .models import MeasurementLog
from .models import CableInventoryItem


class CableCommission(TemplateExtension):
    model = 'dcim.cable'

    def buttons(self):
        cable = self.context['object']
        log = None

        maybe_log = MeasurementLog.objects.filter(cable=cable)
        if maybe_log.exists():
            log = maybe_log.first().link

        cable_inventory_item = CableInventoryItem.objects.filter(cable=cable)

        return self.render('nautobot_cable_utils/inc/buttons.html', {
            'cable': cable,
            'cable_planned': cable.status.name == "Planned",
            'cable_inventory': cable_inventory_item.exists(),
            'log': log,
        })


class DeviceBulkConnect(TemplateExtension):
    model = 'dcim.device'

    def buttons(self):
        device = self.context['object']
        cable_available = None

        if device.get_cables().filter(status__name="Planned").count() > 0:
            cable_available = device.get_cables().filter(status__name="Planned").first().pk

        return self.render('nautobot_cable_utils/inc/buttons_device.html', {
            'device': device,
            'cable_planned': cable_available is not None,
        })


class InterfaceAutoRouteStart(TemplateExtension):
    model = "dcim.interface"

    def buttons(self):
        return self.render(
            'nautobot_cable_utils/inc/buttons_interface.html',
        )


class FrontPortAutoRouteStart(TemplateExtension):
    model = "dcim.frontport"

    def buttons(self):
        return self.render(
            'nautobot_cable_utils/inc/buttons_frontport.html',
        )


template_extensions = [CableCommission, DeviceBulkConnect, InterfaceAutoRouteStart, FrontPortAutoRouteStart]



# Generated by Django 3.2.15 on 2022-10-07 07:43

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('nautobot_cable_utils', '0003_rename_tenant_cabletemplate_owner'),
    ]

    operations = [
        migrations.RenameField(
            model_name='cabletemplate',
            old_name='cable_number',
            new_name='name',
        ),
    ]



# Generated by Django 3.2.25 on 2024-06-21 14:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dcim', '0058_controller_data_migration'),
        ('tenancy', '0009_update_all_charfields_max_length_to_255'),
        ('nautobot_cable_utils', '0009_auto_20230802_0724'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='CableTemplate',
            new_name='CableInventoryItem',
        ),
    ]



from importlib import metadata

from nautobot.apps import NautobotAppConfig

__version__ = metadata.version(__name__)


class CableUtilsConfig(NautobotAppConfig):
    name = "nautobot_cable_utils"
    verbose_name = "Cable Utilities"
    description = "A cable utilities plugin"
    version = __version__
    author = "Gesellschaft für wissenschaftliche Datenverarbeitung mbH Göttingen"
    author_email = "netzadmin@gwdg.de"
    base_url = "nautobot-cable-utils"
    min_version = "2.0.0"
    max_version = "2.9999"


config = CableUtilsConfig



from nautobot.dcim.api.serializers import CableSerializer, ManufacturerSerializer
from nautobot.tenancy.api.serializers import TenantSerializer
from rest_framework.serializers import ModelSerializer

from nautobot.apps.api import BaseModelSerializer, CustomFieldModelSerializerMixin
from rest_framework import serializers

from nautobot_cable_utils.models import CableInventoryItem, CablePlug


class NestedCablePlugSerializer(BaseModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="plugins-api:nautobot_cable_utils-api:cableplug-detail")

    class Meta:
        model = CablePlug
        fields = ["id", "url", "name"]


class CablePlugSerializer(ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="plugins-api:nautobot_cable_utils-api:cableplug-detail")

    class Meta:
        model = CablePlug
        fields = [
            "id",
            "url",
            "name",
        ]


class CableInventoryItemSerializer(CustomFieldModelSerializerMixin, BaseModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="plugins-api:nautobot_cable_utils-api:cableinventoryitem-detail")
    owner = TenantSerializer()
    cable = CableSerializer()
    plug = CablePlugSerializer()
    supplier = ManufacturerSerializer()

    class Meta:
        model = CableInventoryItem
        fields = [
            "id",
            "url",
            "name",
            "label",
            "type",
            "plug",
            "color",
            "supplier",
            "procurement_ident",
            "length",
            "length_unit",
            "cable",
            "owner",
        ]



from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from nautobot.extras.models import Status
from nautobot.dcim.choices import CableTypeChoices, InterfaceTypeChoices
from nautobot.dcim.models import Location, LocationType, Manufacturer, DeviceType, Device, Rack, RearPortTemplate, Cable, \
    RearPort, InterfaceTemplate, FrontPortTemplate
from nautobot.extras.models import Role
from nautobot_cable_utils.cable_router import CableRouter


class CableRouterTest(TestCase):

    @classmethod
    def _populate_locations(cls):

        location_type = LocationType.objects.create(name="Site")
        status = Status.objects.get(name="Active")
        location = Location.objects.create(name="1519 - RZGö", location_type=location_type, status=status)
        manufacturer = Manufacturer.objects.create(name="Manufacturer 1")
        dt_pp = DeviceType.objects.create(manufacturer=manufacturer, model="Patch Panel")

        rpt = RearPortTemplate.objects.create(
            device_type=dt_pp,
            name="1/1"
        )

        FrontPortTemplate.objects.create(
            device_type=dt_pp,
            name="1/1",
            rear_port_template=rpt,
        )

        dt_d = DeviceType.objects.create(manufacturer=manufacturer, model="Device")

        InterfaceTemplate.objects.create(
            device_type=dt_d,
            name="swp1",
            type=InterfaceTypeChoices.TYPE_100GE_QSFP28,
        )
        device_content_type = ContentType.objects.get_for_model(Device)
        role = Role.objects.create(name="Device Role 1")
        role.content_types.set([device_content_type])

        return location, manufacturer, dt_pp, dt_d, role

    @classmethod
    def _populate_edges(cls, location, dt_pp, dt_d, role, edges, device_locations, cable_type_choice):

        rack_names = set([x for e in edges for x in e])
        status = Status.objects.get(name="Active")
        for rack_name in rack_names:
            Rack.objects.create(
                name=rack_name,
                location=location,
                status=status,
            )

        patched_cables = list()

        for edge in edges:
            (rack_a, rack_b) = edge

            device_a = Device.objects.create(
                role=role,
                device_type=dt_pp,
                location=location,
                rack=Rack.objects.get(location=location, name=rack_a),
                status=status,
            )

            device_b = Device.objects.create(
                role=role,
                device_type=dt_pp,
                location=location,
                rack=Rack.objects.get(location=location, name=rack_b),
                status=status,
            )

            c = Cable.objects.create(
                termination_a=RearPort.objects.get(device=device_a),
                termination_b=RearPort.objects.get(device=device_b),
                type=cable_type_choice,
                status=status,
            )

            patched_cables.append(c)

        start_dev = Device.objects.create(
            role=role,
            device_type=dt_d,
            location=location,
            rack=Rack.objects.get(location=location, name=device_locations[0]),
            status=status,
        )

        end_dev = Device.objects.create(
            role=role,
            device_type=dt_d,
            location=location,
            rack=Rack.objects.get(location=location, name=device_locations[1]),
            status=status,
        )

        return patched_cables, start_dev, end_dev

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        location, manufacturer, dt_pp, dt_d, role = cls._populate_locations()
        cls.simple_patch_cables, cls.simple_patch_start, cls.simple_patch_end = cls._populate_edges(
            location, dt_pp, dt_d, role,
            [
                ("1519-B-1009", "1519-B-1010"),
                ("1519-B-1011", "1519-B-1012"),
                ("1519-B-1010", "1519-B-1011"),
            ],
            [
                "1519-B-1009",
                "1519-B-1012",
            ],
            CableTypeChoices.TYPE_SMF
        )

        cls.neighbour_patch_cables, cls.neighbour_patch_start, cls.neighbour_patch_end = cls._populate_edges(
            location, dt_pp, dt_d, role,
            [
                ("1519-A-1009", "1519-A-1010"),
            ],
            [
                "1519-A-1009",
                "1519-A-1010",
            ],
            CableTypeChoices.TYPE_SMF
        )

        cls.rack_hop_cables, cls.rack_hop_start, cls.rack_hop_end = cls._populate_edges(
            location, dt_pp, dt_d, role,
            [
                ("1519-C-1009", "1519-C-1010"),
                ("1519-C-1011", "1519-C-1012"),
            ],
            [
                "1519-C-1009",
                "1519-C-1012",
            ],
            CableTypeChoices.TYPE_SMF
        )

    def test_dummy(self):

        content_type = ContentType.objects.get(app_label="dcim", model="interface")
        cr = CableRouter(self.simple_patch_start.interfaces.first(), content_type, self.simple_patch_end.interfaces.first(), content_type, "fiber_sm")
        cr_reverse = CableRouter(self.simple_patch_end.interfaces.first(), content_type, self.simple_patch_start.interfaces.first(), content_type, "fiber_sm")
        cr_simple_hop = CableRouter(self.neighbour_patch_end.interfaces.first(), content_type, self.neighbour_patch_start.interfaces.first(), content_type, "fiber_sm", enable_next_rack_hops=True)
        path = cr.get_path()
        path_reverse = cr_reverse.get_path()
        path_simple_hop = cr_simple_hop.get_path()

        self.assertEqual(len(path), 3)
        self.assertEqual(len(path_reverse), 3)
        self.assertEqual(len(path_simple_hop), 1)
        self.assertEqual(sorted(map(lambda c: c['cable_id'], path)), sorted(map(lambda c: c.id, self.simple_patch_cables)))
        self.assertEqual(sorted(map(lambda c: c['cable_id'], path_reverse)), sorted(map(lambda c: c.id, self.simple_patch_cables)))

    def test_rack_hop_disabled(self):
        content_type = ContentType.objects.get(app_label="dcim", model="interface")
        cr = CableRouter(self.rack_hop_start.interfaces.first(), content_type, self.rack_hop_end.interfaces.first(), content_type, "fiber_sm")
        cr_easy_hop = CableRouter(self.rack_hop_start.interfaces.first(), content_type, self.rack_hop_end.interfaces.first(), content_type, "fiber_sm", enable_next_rack_hops=True)
        path = cr.get_path()
        cr_easy_hop_path = cr_easy_hop.get_path()
        self.assertEqual(path, [])
        self.assertEqual(len(cr_easy_hop_path), 2)



# Generated by Django 3.2.25 on 2024-06-21 14:09

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('dcim', '0058_controller_data_migration'),
        ('tenancy', '0009_update_all_charfields_max_length_to_255'),
        ('nautobot_cable_utils', '0010_rename_cabletemplate_cableinventoryitem'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cableinventoryitem',
            name='owner',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='cable_inventory_items', to='tenancy.tenant'),
        ),
        migrations.AlterField(
            model_name='cableinventoryitem',
            name='plug',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='cable_inventory_items', to='nautobot_cable_utils.cableplug'),
        ),
        migrations.AlterField(
            model_name='cableinventoryitem',
            name='storage_rack',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='cable_inventory_items', to='dcim.rack'),
        ),
        migrations.AlterField(
            model_name='cableinventoryitem',
            name='supplier',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='cable_inventory_items', to='dcim.manufacturer'),
        ),
    ]



from nautobot.apps.ui import ButtonColorChoices, NavMenuTab, NavMenuGroup, NavMenuItem, NavMenuButton


menu_items = (
    NavMenuTab(
        name="Plugins",
        groups=[
            NavMenuGroup(name="Cable Utilities", items=[
                NavMenuItem(
                    link='plugins:nautobot_cable_utils:cableinventoryitem_list',
                    name="Cable Inventory",
                    permissions=["dcim.view_cables"],
                    buttons=[
                        NavMenuButton(
                            link='plugins:nautobot_cable_utils:cableinventoryitem_add',
                            title="Add Cable Inventory Item",
                            icon_class="mdi mdi-plus-thick",
                            button_class=ButtonColorChoices.GREEN
                        ),
                        NavMenuButton(
                            link='plugins:nautobot_cable_utils:cableinventoryitem_import',
                            title="Import Cable Inventory Items",
                            icon_class="mdi mdi-database-import-outline",
                            button_class=ButtonColorChoices.BLUE
                        ),
                    ]
                ),
                NavMenuItem(
                    link='plugins:nautobot_cable_utils:measurement_log_list',
                    name="Measurement Logs",
                    permissions=["dcim.view_cables"],
                    buttons=[
                        NavMenuButton(
                            link='plugins:nautobot_cable_utils:measurement_log_add',
                            title="Add Measurement Log",
                            icon_class="mdi mdi-plus-thick",
                            button_class=ButtonColorChoices.GREEN
                        ),
                    ]
                ),
            ]),
        ],
    ),
)



import django_filters
from django.db.models import Q

from nautobot.dcim.models import Manufacturer
from nautobot.apps.filters import NautobotFilterSet
from nautobot.tenancy.models import Tenant

from .models import CableInventoryItem, CablePlug


class CableInventoryItemFilterSet(NautobotFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label='Search',
    )

    owner_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Tenant.objects.all(),
        label="Owner (ID)",
    )
    owner = django_filters.ModelMultipleChoiceFilter(
        queryset=Tenant.objects.all(),
        field_name="owner__name",
        to_field_name="name",
        label="Owner (Name)",
    )

    plug_id = django_filters.ModelMultipleChoiceFilter(
        queryset=CablePlug.objects.all(),
        label="Plug (ID)",
    )

    plug = django_filters.ModelMultipleChoiceFilter(
        queryset=CablePlug.objects.all(),
        field_name="plug__name",
        to_field_name="name",
        label="Plug (Name)",
    )

    supplier_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Manufacturer.objects.all(),
        label="Supplier (ID)",
    )

    supplier = django_filters.ModelMultipleChoiceFilter(
        queryset=Manufacturer.objects.all(),
        field_name="supplier__name",
        to_field_name="name",
        label="Supplier (Name)",
    )

    class Meta:
        model = CableInventoryItem
        fields = ['name', 'type', 'color', "procurement_ident"]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        qs_filter = (
                Q(name__icontains=value) |
                Q(label__icontains=value) |
                Q(procurement_ident__icontains=value)
        )
        return queryset.filter(qs_filter)



from django.contrib import admin

from nautobot_cable_utils.models import CablePlug


@admin.register(CablePlug)
class CablePlugAdmin(admin.ModelAdmin):
    fields = ("name",)
    list_display = ("name",)
    ordering = ("name",)
    search_fields = ("name",)



from . import serializers, filters
from nautobot.apps.api import APIRootView, ModelViewSet
from ..models import CableInventoryItem, CablePlug


class SFPInventoryRootView(APIRootView):

    def get_view_name(self):
        return "SFP Inventory"


class CableInventoryItemViewSet(ModelViewSet):
    queryset = CableInventoryItem.objects.all()
    serializer_class = serializers.CableInventoryItemSerializer
    filterset_class = filters.CableInventoryItemFilterSet


class CablePlugViewSet(ModelViewSet):
    queryset = CablePlug.objects.all()
    serializer_class = serializers.CablePlugSerializer
    filterset_class = filters.CablePlugFilterSet



from django.contrib import messages
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.views.generic import View, FormView

from nautobot.apps.views import BulkCreateView, BulkDeleteView, BulkEditView, BulkImportView, GetReturnURLMixin, ObjectEditView, ObjectListView
from nautobot.dcim.models import Cable, Device, Interface, FrontPort
from nautobot.apps.forms import ConfirmationForm
from nautobot.extras.models.statuses import Status

from . import filters, forms, tables
from .cable_router import CablePath
from .forms import CableRouterVerifyForm, CableRouterLinkToInterfaceForm, \
    CableRouterLinkToFrontPortForm
from .models import CableInventoryItem, MeasurementLog
from .tables import NeededCableTable


class ReconnectView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = 'dcim.add_cable'
    template_name = 'nautobot_cable_utils/cable_connect.html'

    def dispatch(self, request, *args, pk=None, **kwargs):
        self.obj = Cable.objects.get(pk=pk)

        idx = (self.obj.termination_a_type.model, self.obj.termination_b_type.model)
        idx_sorted = tuple(sorted(idx))
        if idx != idx_sorted:
            termination_a = self.obj.termination_a
            self.obj.termination_a = self.obj.termination_b
            self.obj.termination_b = termination_a

        self.form_class = {
            ('circuittermination', 'circuittermination'): forms.ConnectCircuitTerminationForm,
            ('circuittermination', 'rearport'): forms.ConnectCircuitTerminationToRearPortForm,
            ('consoleport', 'consoleserverport'): forms.ConnectConsolePortToConsoleServerPortForm,
            ('consoleport', 'frontport'): forms.ConnectConsolePortToFrontPortForm,
            ('consoleport', 'rearport'): forms.ConnectConsolePortToRearPortForm,
            ('consoleserverport', 'frontport'): forms.ConnectConsoleServerPortToFrontPortForm,
            ('consoleserverport', 'rearport'): forms.ConnectConsoleServerPortToRearPortForm,
            ('powerfeed', 'powerport'): forms.ConnectPowerfeedToPowerPortForm,
            ('poweroutlet', 'powerport'): forms.ConnectPowerOutletToPowerPortForm,
            ('circuittermination', 'interface'): forms.ConnectCircuitTerminationToInterfaceForm,
            ('frontport', 'interface'): forms.ConnectFrontPortToInterfaceForm,
            ('interface', 'rearport'): forms.ConnectInterfaceToRearPortForm,
            ('interface', 'interface'): forms.ConnectInterfaceForm,
            ('frontport', 'frontport'): forms.ConnectFrontPortForm,
            ('frontport', 'rearport'): forms.ConnectFrontPortToRearPortForm,
            ('circuittermination', 'frontport'): forms.ConnectCircuitTerminationToFrontPortForm,
            ('rearport', 'rearport'): forms.ConnectRearPortForm,
        }[idx_sorted]

        return super().dispatch(request, *args, **kwargs)

    def prefill_form(self, initial_data, termination):
        o = getattr(self.obj, termination)
        if o and hasattr(o, 'device'):
            device = o.device
            initial_data['{}_device'.format(termination)] = device
            if device.location:
                initial_data['{}_location'.format(termination)] = device.location
            if device.rack:
                initial_data['{}_rack'.format(termination)] = device.rack

    def get(self, request, *args, **kwargs):
        # Parse initial data manually to avoid setting field values as lists
        initial_data = {k: request.GET[k] for k in request.GET}

        self.prefill_form(initial_data, 'termination_a')
        self.prefill_form(initial_data, 'termination_b')

        form = self.form_class(instance=self.obj, initial=initial_data)

        return render(request, self.template_name, {
            'obj': self.obj,
            'obj_type': Cable._meta.verbose_name,
            'form': form,
            'return_url': self.get_return_url(request, self.obj),
        })

    def post(self, request, *args, **kwargs):
        assigned_cable_inventory_item = None

        with transaction.atomic():
            try:
                if hasattr(self.obj.termination_a, '_path') and self.obj.termination_a._path_id:
                    self.obj.termination_a._path = None
            except CablePath.DoesNotExist:
                pass
            try:
                if hasattr(self.obj.termination_b, '_path') and self.obj.termination_b._path_id:
                    self.obj.termination_b._path = None
            except CablePath.DoesNotExist:
                pass

            assigned_cable_inventory_item = CableInventoryItem.objects.filter(cable=self.obj).first()

            self.obj.delete()
            self.obj.pk = None
            self.obj._state.adding = True

            self.obj.termination_a.cable = None
            self.obj.termination_b.cable = None

            form = self.form_class(request.POST, request.FILES, instance=self.obj)
            if form.is_valid():
                obj = form.save()

                if assigned_cable_inventory_item is not None:
                    # restore cable template if previously assigned
                    assigned_cable_inventory_item.cable = obj
                    assigned_cable_inventory_item.save()

                redirect_url = self.get_return_url(request, obj)
                return redirect(redirect_url)

            return render(request, self.template_name, {
                'obj': self.obj,
                'obj_type': Cable._meta.verbose_name,
                'form': form,
                'return_url': self.get_return_url(request, self.obj),
            })


class CommissionView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = 'dcim.edit_cable'
    template_name = 'nautobot_cable_utils/commission_cable.html'

    def get(self, request, pk=None):
        cable = get_object_or_404(Cable, pk=pk)
        form = forms.CommissionForm(initial={'cable': cable})

        return render(request, self.template_name, {
            'form': form,
            'cable': cable,
            'return_url': self.get_return_url(request, cable),
        })

    def post(self, request, pk=None):
        with transaction.atomic():
            cable = get_object_or_404(Cable, pk=pk)
            form = forms.CommissionForm(request.POST, initial={'cable': cable})

            if form.is_valid():
                form.save()

                return redirect(self.get_return_url(request, cable))

            return render(request, self.template_name, {
                'form': form,
                'cable': cable,
                'return_url': self.get_return_url(request, cable),
            })


class UndoCommissionView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = 'dcim.edit_cable'
    queryset = Cable.objects.all()
    template_name = 'nautobot_cable_utils/undo_commission.html'

    def get(self, request, pk=None):
        cable = get_object_or_404(Cable, pk=pk)
        form = ConfirmationForm(initial=request.GET)

        return render(
            request,
            self.template_name,
            {
                "obj": cable,
                "form": form,
                "obj_type": self.queryset.model._meta.verbose_name,
                "return_url": self.get_return_url(request, cable),
            },
        )

    def post(self, request, pk=None):
        cable = get_object_or_404(Cable, pk=pk)
        form = ConfirmationForm(request.POST)

        if form.is_valid():
            with transaction.atomic():
                cable_new = Cable()
                cable_new.termination_a = cable.termination_a
                cable_new.termination_b = cable.termination_b
                cable_new.status = Status.objects.get(name="Planned")

                cable.delete()
                cable_new.save()

        return redirect(self.get_return_url(request, cable_new))


class DeviceCommissionView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = 'dcim.edit_cable'
    template_name = 'nautobot_cable_utils/bulk_commission.html'

    def get(self, request, pk=None):
        device = get_object_or_404(Device, pk=pk)

        # set key, if not present
        if 'current_device' not in request.session:
            request.session['current_device'] = None

        if request.session['current_device'] != str(pk):
            cable_list = list()
            for cable in device.get_cables().filter(status__name="Planned"):
                cable_list.append(str(cable.pk))

            request.session['current_device'] = str(pk)
            request.session['cable_list'] = cable_list
            request.session.modified = True

        cable = get_object_or_404(Cable, pk=request.session['cable_list'][0])
        form = forms.CommissionForm(initial={'cable': cable})

        # ugly patch to circumvent the need to create a contact_association table like in https://github.com/nautobot/nautobot/blob/768a84e3d16fab359b8355e595348defa55e47c9/nautobot/core/views/generic.py#L146
        # TODO: create proper template
        cable.is_contact_associable_model = False
        return render(request, self.template_name, {
            'form': form,
            'count': len(request.session['cable_list']),
            'object': cable,
            'verbose_name': Cable._meta.verbose_name,
            'verbose_name_plural': Cable._meta.verbose_name_plural,
            'return_url': self.get_return_url(request, device),
            'active_tab': 'main',
        })

    def post(self, request, pk=None):
        with transaction.atomic():
            device = get_object_or_404(Device, pk=pk)
            cable = get_object_or_404(Cable, pk=request.session['cable_list'][0])
            form = forms.CommissionForm(request.POST, initial={'cable': cable})

            if '_create' in self.request.POST:
                if form.is_valid():
                    request.session['cable_list'].pop(0)
                    request.session.modified = True
                    form.save()
            elif '_skip' in self.request.POST:
                request.session['cable_list'].pop(0)
                request.session.modified = True

            if len(request.session['cable_list']) > 0 and '_cancel' not in self.request.POST:
                if form.is_valid() or '_skip' in self.request.POST:
                    return redirect(request.path)
                else:
                    # re-render page to display ValidationErrors
                    return render(request, self.template_name, {
                        'form': form,
                        'count': len(request.session['cable_list']),
                        'object': get_object_or_404(Cable, pk=request.session['cable_list'][0]),
                        'verbose_name': Cable._meta.verbose_name,
                        'verbose_name_plural': Cable._meta.verbose_name_plural,
                        'return_url': self.get_return_url(request, device),
                        'active_tab': 'main',
                    })

            if 'current_device' in self.request.session.keys():
                del request.session['current_device']
            if 'cable_list' in self.request.session.keys():
                del request.session['cable_list']

        return redirect(self.get_return_url(request, device))


class CableInventoryItemListView(ObjectListView):
    queryset = CableInventoryItem.objects.all()
    filterset = filters.CableInventoryItemFilterSet
    filterset_form = forms.CableInventoryItemFilterForm
    table = tables.CableInventoryItemTable


class CableInventoryItemCreateView(ObjectEditView):
    queryset = CableInventoryItem.objects.all()
    model_form = forms.CableInventoryItemForm
    default_return_url = 'plugins:nautobot_cable_utils:cableinventoryitem_list'
    template_name = "nautobot_cable_utils/cableinventoryitem_edit.html"


class CableInventoryItemBulkCreateView(BulkCreateView):
    queryset = CableInventoryItem.objects.all()
    form = forms.CableInventoryItemBulkCreateForm
    model_form = forms.CableInventoryItemBulkForm
    pattern_target = "name"
    template_name = "nautobot_cable_utils/cableinventoryitem_bulk_add.html"


class CableInventoryItemBulkEditView(BulkEditView):
    queryset = CableInventoryItem.objects.all()
    form = forms.CableInventoryItemBulkEditForm
    model_form = forms.CableInventoryItemBulkForm
    table = tables.CableInventoryItemTable


class CableInventoryItemEditView(CableInventoryItemCreateView):
    pass


class CableInventoryItemBulkImportView(BulkImportView):
    queryset = CableInventoryItem.objects.all()
    model_form = forms.CableInventoryItemCSVForm
    table = tables.CableInventoryItemTable


class CableInventoryItemBulkDeleteView(BulkDeleteView):
    queryset = CableInventoryItem.objects.all()
    filterset = filters.CableInventoryItemFilterSet
    table = tables.CableInventoryItemTable


class MeasurementLogListView(ObjectListView):
    queryset = MeasurementLog.objects.all()
    table = tables.MeasurementLogTable
    action_buttons = tuple()
    template_name = 'nautobot_cable_utils/measurement_log_list.html'


class MeasurementLogCreateView(ObjectEditView):
    queryset = MeasurementLog.objects.all()
    model_form = forms.MeasurementLogForm
    default_return_url = 'plugins:nautobot_cable_utils:measurement_log_list'


class MeasurementLogEditView(CableInventoryItemCreateView):
    pass


class MeasurementLogBulkDeleteView(BulkDeleteView):
    queryset = MeasurementLog.objects.all()
    table = tables.MeasurementLogTable


class CableRouterAddView(PermissionRequiredMixin, FormView):
    permission_required = 'dcim.add_cable'
    template_name = "nautobot_cable_utils/cable_router_add.html"

    termination_types = {
        "interface": ContentType.objects.get_for_model(Interface),
        "front-port": ContentType.objects.get_for_model(FrontPort),
    }

    termination_forms = {
        "interface": CableRouterLinkToInterfaceForm,
        "front-port": CableRouterLinkToFrontPortForm,
    }

    termination_names = {
        'interface': "Interface",
        "front-port": "Front Port",
    }

    termination_a_type = None
    termination_b_type = None
    termination_a_type_name = None
    termination_b_type_name = None
    termination_a = None
    form_class = None

    def dispatch(self, request, *args, **kwargs):
        self.termination_a_type = self.termination_types[kwargs.get("termination_a_type")]
        self.termination_b_type = self.termination_types[kwargs.get("termination_b_type")]
        self.termination_a_type_name = self.termination_names[kwargs.get("termination_a_type")]
        self.termination_b_type_name = self.termination_names[kwargs.get("termination_b_type")]
        self.termination_a = self.termination_a_type.model_class().objects.get(id=kwargs.get("termination_a_id"))
        self.form_class = self.termination_forms[kwargs.get("termination_b_type")]

        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        return {
            **kwargs,
            "termination_a": self.termination_a,
            "termination_a_type": self.termination_a_type,
            "termination_b_type": self.termination_b_type,
        }

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data()

        return {
            **ctx,
            "termination_a_type": self.termination_a_type,
            "termination_b_type": self.termination_b_type,
            "termination_a_type_name": self.termination_a_type_name,
            "termination_b_type_name": self.termination_b_type_name,
            "termination_a": self.termination_a,
        }

    def form_valid(self, form):
        query_params1 = map(lambda x: f"cable={x['cable_id']}_{'r' if x['is_reversed'] else 'n'}",
                            form.cleaned_data['path'])
        query_params2 = [
            f"media_type={form.cleaned_data['media_type']}",
            f"termination_a_type={self.termination_a_type.id}",
            f"termination_a_id={self.termination_a.id}",
            f"termination_b_type={self.termination_b_type.id}",
            f"termination_b_id={form.cleaned_data['termination_b_id']}"
        ]
        query_params = "&".join([*query_params1, *query_params2])
        redirect_url = reverse("plugins:nautobot_cable_utils:cable_router_verify") + "?" + query_params
        return HttpResponseRedirect(redirect_url)


class CableRouterVerifyView(PermissionRequiredMixin, FormView):
    permission_required = 'dcim.add_cable'
    form_class = CableRouterVerifyForm
    template_name = "nautobot_cable_utils/cable_router_verify.html"

    def get_cable_path(self):
        ct_termination_a = ContentType.objects.get_for_id(self.request.GET.get("termination_a_type"))
        ct_termination_b = ContentType.objects.get_for_id(self.request.GET.get("termination_b_type"))

        termination_a = ct_termination_a.model_class().objects.get(pk=self.request.GET.get("termination_a_id"))
        termination_b = ct_termination_b.model_class().objects.get(pk=self.request.GET.get("termination_b_id"))

        cable_list_raw = self.request.GET.getlist("cable")
        cable_path = CablePath(termination_a, ct_termination_a, termination_b, ct_termination_b, cable_list_raw)

        return cable_path

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        cable_path = self.get_cable_path()

        needed_cable_table = NeededCableTable(cable_path.needed_cables)

        return {
            **ctx,
            'termination_a': cable_path.termination_a,
            "termination_b": cable_path.termination_b,
            "path": cable_path.get_renderable_path(),
            "needed_cable_table": needed_cable_table,
        }

    def form_valid(self, form):
        cable_path = self.get_cable_path()
        needed_cables = cable_path.needed_cables
        planned_status = Status.objects.get(name="Planned")

        for needed_cable in needed_cables:
            Cable.objects.create(
                termination_a=needed_cable['termination_a'],
                termination_b=needed_cable['termination_b'],
                status=planned_status,
            )

        messages.success(self.request, f"{len(needed_cables)} Cables were created.")

        return HttpResponseRedirect(cable_path.termination_a.get_absolute_url())



# Generated by Django 3.2.18 on 2023-03-24 10:55

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tenancy', '0002_auto_slug'),
        ('dcim', '0019_device_redundancy_group_data_migration'),
        ('nautobot_cable_utils', '0007_alter_cabletemplate_supplier'),
    ]

    operations = [
        migrations.AddField(
            model_name='cabletemplate',
            name='storage_rack',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='cable_templates', to='dcim.rack'),
        ),
        migrations.AlterField(
            model_name='cabletemplate',
            name='owner',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='cable_templates', to='tenancy.tenant'),
        ),
    ]



import re

from django.contrib.contenttypes.models import ContentType

from igraph import Graph
from nautobot.dcim.choices import CableTypeChoices
from nautobot.dcim.models import RearPort, Cable, Rack


def get_weight_for_frontport_name(fp_name):
    if "/" in fp_name:
        [part_1, part_2] = fp_name.split("/")
        part_1_match = re.search(r'\d+', part_1)
        part_2_match = re.search(r'\d+', part_2)

        part_1_num = int(part_1_match.group(0)) if part_1_match else 0
        part_2_num = int(part_2_match.group(0)) if part_2_match else 0
        return 1000000 + (part_1_num * 1000) + part_2_num

    else:
        part_1_match = re.search(r'\d+', fp_name)
        part_1_num = int(part_1_match.group(0)) if part_1_match else 0
        return 1000000 + part_1_num


class CablePathExistingMember:
    cable: Cable = None
    is_swapped: bool = False

    def __init__(self, cable_list_raw_entry):
        # cable_list_raw_entry might me something like r_UUID or n_UUID, which means the cable with the UUID in normal
        # or reversed orientation.

        cable_match_regex = re.match(
            r'^([0-9a-fA-F]{8}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{12})_([rn])$',
            cable_list_raw_entry)

        self.cable = Cable.objects.get(pk=cable_match_regex.group(1))
        self.is_swapped = cable_match_regex.group(2) == "r"

        # TODO: Add assertions about connected terminations on cable.

    """
    frontport is an abstraction about the Netbox data model.
    Cables added to CablePathExistingMember are connected to two rear ports.
    Those rear ports should only have one front port.
    """

    @property
    def frontport_a(self):
        return self.cable.termination_b.front_ports.first() if self.is_swapped else \
            self.cable.termination_a.front_ports.first()

    @property
    def frontport_b(self):
        return self.cable.termination_a.front_ports.first() if self.is_swapped else \
            self.cable.termination_b.front_ports.first()

    @property
    def rearport_a(self):
        return self.cable.termination_b if self.is_swapped else self.cable.termination_a

    @property
    def rearport_b(self):
        return self.cable.termination_a if self.is_swapped else self.cable.termination_b


class CablePathNewMember:
    # termination_a and termination_b are swapped correctly beforehand.
    # There is no way that those need to be swapped at any point.
    termination_a = None
    termination_b = None

    def __init__(self, termination_a, termination_b):
        self.termination_a = termination_a
        self.termination_b = termination_b

    @property
    def needed_cable(self):
        return {
            "termination_a_type": ContentType.objects.get_for_model(self.termination_a),
            "termination_a_id": self.termination_a.id,
            "termination_a": self.termination_a,
            "termination_b_type": ContentType.objects.get_for_model(self.termination_b),
            "termination_b_id": self.termination_b.id,
            "termination_b": self.termination_b
        }


class CablePath:
    def __init__(self, termination_a, termination_a_type, termination_b, termination_b_type, cable_list_raw):
        self.termination_a = termination_a
        self.termination_b = termination_b
        self.termination_a_type = termination_a_type
        self.termination_b_type = termination_b_type

        cp_existing_members = list()
        for clre in cable_list_raw:
            cp_existing_members.append(
                CablePathExistingMember(clre)
            )

        self.cp_existing_members = cp_existing_members
        self.cp_members = list()

        open_elem = self.termination_a

        for existing_member in self.cp_existing_members:

            # open_elem might be the same as frontport_a, if we started our search from a front_port and this
            # is the first cable. Thus, we do not need a cable.
            if open_elem != existing_member.frontport_a:
                self.cp_members.append(
                    CablePathNewMember(
                        open_elem,
                        existing_member.frontport_a
                    )
                )

            open_elem = existing_member.frontport_b
            self.cp_members.append(existing_member)

        if open_elem and open_elem != self.termination_b:
            self.cp_members.append(
                CablePathNewMember(
                    open_elem,
                    self.termination_b
                )
            )

    @property
    def needed_cables(self):
        return list(
            map(lambda x: x.needed_cable, filter(lambda x: isinstance(x, CablePathNewMember), self.cp_members)))

    def get_renderable_path(self):

        # needed_cables and self.cable_list are sorted and only needs to be zipped for a complete cable trace.
        # Furthermore, we only need to add interfaces and devices for a neat trace.

        intermediate_elements = list()

        open_element = {
            "device": self.termination_a.device,
        }

        for cpm in self.cp_members:
            if isinstance(cpm, CablePathNewMember):
                intermediate_elements.append({
                    **open_element,
                    "attachment_b": cpm.termination_a
                })
                intermediate_elements.append({
                    "needed_cable": cpm.needed_cable
                })
                open_element = {
                    "attachment_a": cpm.termination_b,
                    "device": cpm.termination_b.device
                }
            elif isinstance(cpm, CablePathExistingMember):
                intermediate_elements.append({
                    **open_element,
                    "attachment_b": cpm.rearport_a
                })
                intermediate_elements.append({
                    "cable": cpm.cable
                })
                open_element = {
                    "attachment_a": cpm.rearport_b,
                    "device": cpm.rearport_b.device
                }

        intermediate_elements.append(open_element)

        return intermediate_elements


class CableRouter:

    def __init__(self, termination_a, termination_a_type, termination_b, termination_b_type, media_type,
                 enable_next_rack_hops=False):
        self.termination_a = termination_a
        self.termination_b = termination_b

        self.allowed_cable_types = None
        if media_type == "fiber_sm":
            self.allowed_cable_types = [
                CableTypeChoices.TYPE_SMF,
                CableTypeChoices.TYPE_SMF_OS1,
                CableTypeChoices.TYPE_SMF_OS2,
            ]
        elif media_type == "fiber_mm":
            self.allowed_cable_types = [
                CableTypeChoices.TYPE_MMF,
                CableTypeChoices.TYPE_MMF_OM1,
                CableTypeChoices.TYPE_MMF_OM2,
                CableTypeChoices.TYPE_MMF_OM3,
                CableTypeChoices.TYPE_MMF_OM4,
            ]
        elif media_type == "copper":
            self.allowed_cable_types = [
                CableTypeChoices.TYPE_CAT3,
                CableTypeChoices.TYPE_CAT5,
                CableTypeChoices.TYPE_CAT5E,
                CableTypeChoices.TYPE_CAT6,
                CableTypeChoices.TYPE_CAT6A,
                CableTypeChoices.TYPE_CAT7,
                CableTypeChoices.TYPE_CAT7A,
                CableTypeChoices.TYPE_CAT8,
            ]

        excluded_cable_ids = list()
        if termination_a_type.app_label == "dcim" and termination_a_type.model == "frontport":
            rps_to_exclude = RearPort.objects.filter(device=self.termination_a.device)
            for rp in rps_to_exclude:
                if termination_a.rear_port != rp and rp.cable:
                    excluded_cable_ids.append(rp.cable.id)

        if termination_b_type.app_label == "dcim" and termination_b_type.model == "frontport":
            rps_to_exclude = RearPort.objects.filter(device=self.termination_b.device)
            for rp in rps_to_exclude:
                if termination_b.rear_port != rp and rp.cable:
                    excluded_cable_ids.append(rp.cable.id)

        rack_connection_cables = Cable.objects.raw(
            '''
                SELECT
                    c.*, d_a.rack_id as termination_a_rack_id, d_b.rack_id as termination_b_rack_id,
                    fp_a.name as termination_a_name, fp_b.name as termination_b_name
                FROM dcim_cable c
                LEFT JOIN django_content_type ct_a ON ct_a.id = termination_a_type_id
                LEFT JOIN django_content_type ct_b ON ct_b.id = termination_b_type_id
                LEFT JOIN dcim_rearport rp_a ON rp_a.id = termination_a_id
                LEFT JOIN dcim_rearport rp_b ON rp_b.id = termination_b_id
                INNER JOIN dcim_frontport fp_a ON fp_a.rear_port_id = rp_a.id AND fp_a.cable_id IS NULL
                INNER JOIN dcim_frontport fp_b ON fp_b.rear_port_id = rp_b.id AND fp_b.cable_id IS NULL
                LEFT JOIN dcim_device d_a ON rp_a.device_id = d_a.id
                LEFT JOIN dcim_device d_b ON rp_b.device_id = d_b.id
                WHERE (
                    c.type = ANY(%s) AND
                    NOT c.id = ANY(%s) AND
                    ct_a.app_label = 'dcim' AND ct_a.model = 'rearport' AND
                    ct_b.app_label = 'dcim' AND ct_b.model = 'rearport' AND
                    fp_a.cable_id IS NULL and fp_b.cable_id IS NULL
                )
            ''',
            [self.allowed_cable_types, excluded_cable_ids]
        )

        """
        We have two different types of rack interconnection, which have some constraints to them:
        1. Neighbor-Connections, which can only be used once in a row, therefore you cannot hop from rack to rack.
        2. Rack-Connections, which can be used unlimited times and have no further restrictions, but should
        be avoided, if there is a Neighbor Connection available.

        Therefore, we need to use a DAG.
        Each rack has two nodes, one node named X_Incoming and one node named X_Outgoing.
        Each connection with two plugs, e.g. an same rack or inter rack connection, goes from X_Incoming to Y_Outgoing.
        Each connection between two ports, e.g. a fiber connection through the data center, goes from X_Outgoing to Y_Incoming.
        Note: Each connection adds two edges to the graph, X_Outgoing to Y_Incoming and Y_Outgoing to X_Incoming.
        """

        self.graph = Graph(directed=True)

        self.resolvable_edges = dict()

        racks = Rack.objects.all()

        for rack in racks:
            incoming_vertex_id = f'incoming_{rack.id}'
            outgoing_vertex_id = f'outgoing_{rack.id}'

            self.graph.add_vertex(incoming_vertex_id, human_name=rack.name)
            self.graph.add_vertex(outgoing_vertex_id, human_name=rack.name)
            self.graph.add_edge(incoming_vertex_id, outgoing_vertex_id, weight=0)

        if enable_next_rack_hops:
            rack_tuples = Rack.objects.raw(
                '''SELECT 1 as id, racks1.id as rack1_id, racks2.id as rack2_id FROM dcim_rack as racks1, dcim_rack as racks2 WHERE racks1.location_id=racks2.location_id and is_neighbor(racks1.name, racks2.name)'''
            )
            for rack_tuple in rack_tuples:
                self.graph.add_edge(f"incoming_{rack_tuple.rack1_id}", f"outgoing_{rack_tuple.rack2_id}", weight=1500000)

        for edge in rack_connection_cables:
            fp_a_weight = get_weight_for_frontport_name(edge.termination_a_name)
            fp_b_weight = get_weight_for_frontport_name(edge.termination_b_name)

            if not edge.termination_a_rack_id or not edge.termination_b_rack_id:
                continue

            termination_a_incoming_vertex_id = f'incoming_{edge.termination_a_rack_id}'
            termination_a_outgoing_vertex_id = f'outgoing_{edge.termination_a_rack_id}'
            termination_b_incoming_vertex_id = f'incoming_{edge.termination_b_rack_id}'
            termination_b_outgoing_vertex_id = f'outgoing_{edge.termination_b_rack_id}'

            e_to = self.graph.add_edge(
                termination_a_outgoing_vertex_id,
                termination_b_incoming_vertex_id,
                weight=fp_a_weight,
            )
            self.resolvable_edges[e_to.index] = {
                "cable_id": edge.id,
                "is_reversed": False,
            }
            e_from = self.graph.add_edge(
                termination_b_outgoing_vertex_id,
                termination_a_incoming_vertex_id,
                weight=fp_b_weight,
            )
            self.resolvable_edges[e_from.index] = {
                "cable_id": edge.id,
                "is_reversed": True,
            }

    def get_path(self):

        # Technically, this could also end on incoming_, but there is a zero cost edge in place.
        shortest_path = self.graph.get_shortest_paths(
            f'incoming_{self.termination_a.device.rack.id}',
            f'outgoing_{self.termination_b.device.rack.id}',
            weights="weight",
            output="epath",
        )

        # shortest_path could contain more than one element, if the function is called in a different way.
        # Therefore, we can unpack this.

        if len(shortest_path) < 1:
            return None

        # If res now contains no elements, this is a direct connection, which can be done without any "real" edges.
        res = [self.resolvable_edges[x] for x in shortest_path[0] if x in self.resolvable_edges]
        return res



import django_filters
from django.db.models import Q

from nautobot_cable_utils.models import CableInventoryItem, CablePlug
from nautobot.apps.filters import BaseFilterSet, CreatedUpdatedModelFilterSetMixin


class CablePlugFilterSet(
    BaseFilterSet,
    CreatedUpdatedModelFilterSetMixin,
):
    q = django_filters.CharFilter(
        method="search",
        label="Search",
    )

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(Q(name__icontains=value))

    class Meta:
        model = CablePlug
        fields = ["id", "name"]


class CableInventoryItemFilterSet(
    BaseFilterSet,
    CreatedUpdatedModelFilterSetMixin,
):
    q = django_filters.CharFilter(
        method="search",
        label="Search",
    )

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value) | Q(label__icontains=value)
        )

    class Meta:
        model = CableInventoryItem
        fields = [
            "id",
            "name",
            "label",
            "type",
            "plug",
            "color",
            "supplier",
            "procurement_ident",
            "length",
            "length_unit",
            "cable",
            "owner",
        ]



from nautobot.apps.api import OrderedDefaultRouter
from .views import CablePlugViewSet, CableInventoryItemViewSet

router = OrderedDefaultRouter()
router.register("cable-plug", CablePlugViewSet)
router.register("cable-inventory-item", CableInventoryItemViewSet)

app_name = "nautobot_cable_utils-api"
urlpatterns = router.urls



from django.db import models

from nautobot.dcim.choices import CableTypeChoices, CableLengthUnitChoices
from nautobot.dcim.models import Cable
from nautobot.apps.models import ColorField, RestrictedQuerySet, BaseModel, RelationshipModel


class CablePlug(BaseModel):
    name = models.CharField(
        max_length=50,
        unique=True,
    )

    def __str__(self):
        return self.name


class CableInventoryItem(BaseModel, RelationshipModel):
    name = models.CharField(
        max_length=100,
        unique=True,
    )
    label = models.CharField(
        max_length=100,
        blank=True
    )
    type = models.CharField(
        max_length=50,
        choices=CableTypeChoices,
        blank=True
    )
    plug = models.ForeignKey(
        to=CablePlug,
        on_delete=models.CASCADE,
        related_name="cable_inventory_items",
        blank=True,
        null=True,
    )
    color = ColorField(
        blank=True
    )
    length = models.PositiveSmallIntegerField(
        blank=True,
        null=True
    )
    length_unit = models.CharField(
        max_length=50,
        choices=CableLengthUnitChoices,
        blank=True,
    )
    cable = models.OneToOneField(
        Cable,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    owner = models.ForeignKey(
        to="tenancy.Tenant",
        on_delete=models.PROTECT,
        related_name="cable_inventory_items",
        blank=True,
        null=True,
    )

    supplier = models.ForeignKey(
        to="dcim.Manufacturer",
        on_delete=models.PROTECT,
        related_name="cable_inventory_items",
        null=True,
        blank=True,
    )

    storage_rack = models.ForeignKey(
        to="dcim.Rack",
        on_delete=models.PROTECT,
        related_name="cable_inventory_items",
        blank=True,
        null=True,
    )

    procurement_ident = models.CharField(
        max_length=100,
        null=True,
        blank=True,
        verbose_name="Procurement Identifier"
    )

    objects = RestrictedQuerySet.as_manager()

    import_csv_headers = [
        'name', 'owner', 'type', 'plug', 'label', 'color', 'length', 'length_unit', "storage_rack", "supplier",
        "procurement_ident"
    ]

    csv_headers = [
        *import_csv_headers,
        "in_use"
    ]

    def __str__(self):
        return self.name

    def to_csv(self):
        return (
            self.name,
            self.owner.name if self.owner else None,
            self.get_type_display() if self.type else None,
            self.plug.name if self.plug else None,
            self.label,
            self.color,
            self.length,
            self.length_unit,
            self.storage_rack.name if self.storage_rack else None,
            self.supplier.name if self.supplier else None,
            self.procurement_ident,
            self.cable is not None,
        )


class MeasurementLog(BaseModel):
    link = models.URLField(
        blank=True,
        null=True
    )
    cable = models.OneToOneField(
        Cable,
        on_delete=models.CASCADE,
    )



# Generated by Django 3.2.20 on 2023-08-02 07:24

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('nautobot_cable_utils', '0008_auto_20230324_1055'),
    ]

    operations = [
        migrations.RunSQL(
            sql="""
                create or replace function cast_to_int(text, integer) returns integer as $$
                begin
                    return cast($1 as integer);
                exception
                    when invalid_text_representation then
                        return $2;
                end;
                $$ language plpgsql immutable;
                CREATE OR REPLACE FUNCTION get_numberscheme (rack varchar) RETURNS varchar AS $$ BEGIN
                    return case
                        when rack like '1519-%' then 'grz'
                        when rack like 'V1519.%' then 'goenet'
                        when rack like 'V2356.%' then 'goenet'
                        else 'unknown'
                    end;
                END;
                $$ LANGUAGE plpgsql;
                CREATE OR REPLACE FUNCTION is_neighbor (r1 varchar, r2 varchar) RETURNS bool AS $$ DECLARE
                    numberscheme varchar;
                BEGIN
                    numberscheme = get_numberscheme(r1);
                    CASE
                        WHEN numberscheme != get_numberscheme(r2) THEN RETURN false;
                        WHEN numberscheme = 'unknown' THEN RETURN false;
                        WHEN numberscheme = 'goenet' THEN
                            CASE
                                WHEN substr(r1, 1, 8) != substr(r2, 1, 8) THEN RETURN false;
                                WHEN ABS(cast_to_int(substr(r1, 9, 1), 0) - cast_to_int(substr(r2, 9, 1), 0)) = 1 THEN RETURN true;
                                ELSE RETURN false;
                            END CASE;
                        WHEN numberscheme = 'grz' THEN
                            CASE
                                WHEN substr(r1, 1, 9) != substr(r2, 1, 9) THEN RETURN false;
                                WHEN ABS(cast_to_int(substr(r1, 10, 2), 0) - cast_to_int(substr(r2, 10, 2), 0)) = 1 THEN RETURN true;
                                WHEN ABS(cast_to_int(substr(r1, 10, 2), 0) - cast_to_int(substr(r2, 10, 2), 0)) = 2 THEN
                                    CASE
                                        WHEN (
                                            select count(*) from dcim_rack
                                            where name like (
                                                substr(r1, 1, 9) ||
                                                LPAD(CAST(GREATEST(
                                                    cast_to_int(substr(r1, 10, 2), 0)-1,
                                                    cast_to_int(substr(r2, 10, 2), 0)-1
                                                ) as varchar), 2, '0') ||
                                                '%')
                                         ) = 0 THEN RETURN true;
                                        ELSE RETURN false;
                                    END CASE;
                                ELSE RETURN false;
                            END CASE;
                        ELSE
                            RETURN false;
                    END CASE;
                END;
                $$ LANGUAGE plpgsql;
            """
        )
    ]
